// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of mergerays
//
// mergerays is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// mergerays is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <algorithm>

#include <suplibxx/str.h>
#include <tracefctxx/TraceFct.h>

#include "BpipeInput.h"
#include "DataField.h"

BpipeInput::~BpipeInput( ) {
    for ( auto bpipe : bpipes )
        bpipe_delete( bpipe );
}



BpipeInput::BpipeInput( std::string& fnames, const std::vector<std::string>& xfer_fields )
{
    std::vector<std::string> filenames;
    suplib::tok( filenames, fnames, "," );

    attach_bpipes( filenames );

    // Transfer headers to the data field which are requested 
    // with the xfer parameter.
    header_to_datafield( xfer_fields );
}

size_t BpipeInput::read_dpkts( size_t i ) {
    if ( i < 0 || i >= get_num() )
        TraceFct(__func__).die( "Illegal stream id requested" + std::to_string( i ) );

    size_t nr = bpipe_read_dpkts( bpipes[ i ], data[i], 1 );
    if ( nr == 0 && bpipe_errno != BPNOERROR )
        TraceFct(__func__).die( "error reading datapackets from stream id# " + std::to_string(i) +
                                bpipe_strerror( bpipe_errno ) );

    return nr;
}


//
// allocate meory for the bpipe and data members.
//
void
BpipeInput::attach_bpipes( std::vector<std::string>& filenames ) { 

    TraceFct tf( __func__ );
    n_dpkts_valid = 0;
    n_dpkts_sum = 0;
    bpipes.clear();
    data.clear();

    for( auto& filename : filenames ) {

        auto bpipe = bpipe_new();
        if ( bpipe == NULL )
            tf.die( "error allocating bpipe for " + filename );

        if ( bpipe_input( bpipe, filename.c_str() ) )
            tf.die( "error attaching stream to " + filename );

        size_t _n_dpkts = 0;

        // make sure that the header field BPHN_n_dpkts_valid is present.
        int* ptr = static_cast<int*> ( bpipe_hdrf_data( bpipe, BPHN_n_dpkts_valid, 0 ) );

        if ( ptr && 0 != *ptr ) {
            ++n_dpkts_valid;

            size_t* ndpkts_p =
                static_cast<size_t*> (bpipe_hdrf_data( bpipe, BPHN_n_dpkts, 0 ) );

            if ( ! ndpkts_p )
                tf.die( "error in input stream from " + filename
                        + ": n_dpkts_valid is true, but no n_dpkts header field?" );

            _n_dpkts = *ndpkts_p;
        }

        n_dpkts.push_back( _n_dpkts );
        n_dpkts_sum += _n_dpkts;
        bpipes.push_back( bpipe );
    }
}

void
BpipeInput::header_to_datafield( const std::vector<std::string>& xfer_fields ) {

    TraceFct tf( __func__ );

    //
    // Loop over the input stream.
    //
    for ( size_t ii = 0; ii < get_num(); ii++ ) {

        auto bpipe = bpipes[ii];

        void* last = NULL;	             // initialize it for first time through
        const char* name = NULL;
        size_t index;

        // loop until there are no more header fields available
        while ( bpipe_hdrf_next( bpipe, &name, &index, &last ) ) {

            //
            // For each input stream and header.  Loop over the fields of the
            // xfer parameter to see if the current header needs to be transfer
            // to the data field or not.
            //
            bool found = xfer_fields.end() != std::find( xfer_fields.begin(), xfer_fields.end(), name );
            if ( found  ) {
                BPMatrix* matrix = bpipe_hdrf_matrix ( bpipe, name, index );
                BPDataType type = bpipe_hdrf_type( bpipe, name, index );

                if ( NULL == matrix )
                    tf.die( std::string("error allocating matrix for transferring header field (")
                            + name
                            + ") to data packet field:  "
                            + bpipe_strerror( bpipe_errno )
                            );

                switch( bpipe_dpktf_add( bpipe, name, type, matrix ) ) {

                case 1 :
                    tf.die( std::string("error adding header field (")
                            + name
                            + ") to data packet field: field already exists."
                            );
                    break;

                case -1 :
                    tf.die( std::string("error adding header field (")
                            + name
                            + ") to data packet field: "
                            + bpipe_strerror( bpipe_errno )
                            );
                    break;

                default:
                    break;

                }
            }
        }
    }
}

void
BpipeInput::resize_core( std::list<DataFieldPtr>& data_fields, const std::vector<std::string>& xfer_fields ) {

    TraceFct tf( __func__ );

    data.clear();

    size_t num = get_num();
    // Loop over the input streams.
    
    for ( auto& bpipe : bpipes ) {

        // resize the core then map the core.
        for( auto& datafield : data_fields ) {

            BPMatrix*   matrix = datafield->get_matrix( );
            const std::string&  name   = datafield->get_name( );

            DpktField*  dpktf = bpipe_dpktf( bpipe, name.c_str() );

            // if this is a transferred data header field, don't resize it.
            bool found = std::find( xfer_fields.begin(), xfer_fields.end(), name ) != xfer_fields.end();
            if ( ! found ) {
                BPMatrix*   matrix_dup = bpipe_matrix_dup( matrix );
                if ( bpipe_dpktf_resize_core( dpktf, matrix_dup, NULL, NULL, NULL ) )
                    tf.die( "error resizing field: " + name + ": " + bpipe_strerror( bpipe_errno ) );
            }
        }

        data.push_back( bpipe_map_alloc( bpipe, 1, NULL ) );
        if ( ! data.back() )
            tf.die( "error mapping input stream" );
    }

    for ( size_t i = 0; i < num; i++ ) {

        for( auto& datafield : data_fields ) {
            const std::string& name   = datafield->get_name( );
            BPDataType type = datafield->get_type( );

            void* dst = bpipe_dpktf_datap( bpipes[ i ], data[ i ], name.c_str(), type );

            bool found = std::find( xfer_fields.begin(), xfer_fields.end(), name ) != xfer_fields.end();
            if (  found ) {
                BPMatrix*   matrix = datafield->get_matrix( );
                // The index to the call to bpipe_hdrf_data must always be 0
                // since there could only be one field with the given name.
                bpipe_data_copy( dst, bpipe_hdrf_data( bpipes[ i ], name.c_str(), 0 ), matrix, type );
            }
            datafield->set_input_data( i, dst );
        }

    }

}
