// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of mergerays
//
// mergerays is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// mergerays is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include <algorithm>
#include <stdlib.h>
#include <unistd.h>
#include <vector>
#include <list>
#include <cstring>

#include <tracefctxx/TraceFct.h>

#include "BpipeOutput.h"
#include "DataField.h"

int printDataField( void* d ) {
    DataField* ptr = (DataField*) d;
    cout << ptr << endl;
    return 0;
}

BpipeOutput::~BpipeOutput( ) {

    if ( bpipe )
        bpipe_delete( bpipe ); 
}

BpipeOutput::BpipeOutput( BpipeInput& _input, const std::string& _output,
                          const std::vector<std::string>& xfer_fields)
    : input( _input )
    , output( _output )
{
    TraceFct tf( __func__ );
    bpipe = bpipe_new( );
    if ( NULL == bpipe )
        tf.die( "Unable to allocate memory for bpipe" );

    std::list<DataFieldPtr> data_fields;

    // loop through all the input bpipes to write the output header 
    // fields which are not requested to be transferred to the data field.
    //
    for ( size_t i = 0; i < input.get_num( ); i++ ) {
        BPipe* bpipe_input = input.get_bpipe( i );
        add_header_fields( bpipe_input, xfer_fields, data_fields );
    }

    // loop thru all the input bpipes to extract the data fields.
    for ( size_t j = 0; j < input.get_num( ); j++ ) {
        BPipe* bpipe_input = input.get_bpipe( j );
        select_data_fields( bpipe_input, xfer_fields, data_fields );
    }

    if ( input.have_valid_npkts() ) {
        size_t n_dpkts = input.get_n_dpkts_sum();
        size_t n_dpkts_valid = 1;
        if ( bpipe_hdrf_add( bpipe, BPHN_n_dpkts, BPHT_n_dpkts, NULL, &n_dpkts, 1 ) )
            tf.die ( std::string( "error adding header field " BPHN_n_dpkts )
                     + ": "
                     + bpipe_strerror( bpipe_errno ) );

        if ( bpipe_hdrf_add( bpipe, BPHN_n_dpkts_valid, BPHT_n_dpkts_valid, NULL, &n_dpkts_valid, 1 ) )
            tf.die ( std::string( "error adding header field " BPHN_n_dpkts_valid )
                     + ": "
                     + bpipe_strerror( bpipe_errno ) );
    }

    init_bpipe_output( data_fields );
    get_data_field_ptr( data_fields );

    input.resize_core( data_fields, xfer_fields );

    set_output_data_fields( data_fields );
}

void
BpipeOutput::add_header_fields( BPipe* bpipe_input,
                                const std::vector<string>& xfer_fields,
                                std::list<DataFieldPtr>& data_fields ) {

    TraceFct tf( __func__ );
    // last is used by bpipe_hdrf_next to keep track of where it is
    void* last = NULL;	             // initialize it for first time through
    const char* name = NULL;
    size_t index;

    // For a given bpipe, loop until there are no more header fields available
    while ( bpipe_hdrf_next( bpipe_input, &name, &index, &last ) ) {

        // don't add _n_dpkts and _n_dpkts_valid fields.  this should really be
        // handled by the bpipe library, but it isn't.
        if (    0 == std::strcmp( name, BPHN_n_dpkts )
                || 0 == std::strcmp( name, BPHN_n_dpkts_valid )
                )
            continue;
        //
        // extract the data type and a pointer to the data.  these should
        // be valid since we've been handed name and index by bpipe_hdrf_next
        //
        BPDataType type = bpipe_hdrf_type( bpipe_input, name, index );
        void*  thedata = bpipe_hdrf_data( bpipe_input, name, index );

        BPMatrix* matrix = bpipe_hdrf_matrix ( bpipe_input, name, index );

        if ( NULL == matrix )
            tf.die ( std::string( "error allocating header field matrix" ) + bpipe_strerror( bpipe_errno ) );

        // loop thru all the parameters which have
        // to be transferred to the data fields.
        bool wrote_header = false;
        for( auto& xfer_field : xfer_fields ) {

            // Add to the output header fields if it is not
            // requested to be transferred to the data field.
            //
            if ( xfer_field !=  name ) {

                if ( ! wrote_header ) {

                    // The header field is not in the xfer parameter
                    // so write the input stream header field to the
                    // output header.
                    //
                    if ( bpipe_hdrf_add( bpipe, name, type, matrix, thedata, 1 ) )
                        tf.die ( std::string( "error adding header field: " )
                                 + name
                                 + ": "
                                 + bpipe_strerror( bpipe_errno ) );

                    // To make sure that the header field
                    // is only written out once per stream
                    wrote_header = true;
                }

            } else {

                //
                // The header field is to be transferred to the data field.
                //
                select_data_fields( name, type, matrix, data_fields );
            }
        }

    }
}

size_t BpipeOutput::copy_data( size_t i ) {

    size_t n = input.read_dpkts( i );
    if ( 0 == n )
        return 0;

    for( auto& datafield : output_data_fields )
        datafield->copy_data( i );

    return n;
}


void BpipeOutput::get_data_field_ptr( std::list<DataFieldPtr>& data_fields) {

    for( auto& datafield : data_fields ) {

        auto name = datafield->get_name( );
        auto type = datafield->get_type( );

        void* ptr = bpipe_dpktf_datap( bpipe, data, name.c_str(), type );
        if ( NULL == ptr ) 
            TraceFct( __func__ ).die( "no data pointer for data packet field: " + name );

        datafield->set_output_data( ptr );
    }
}

void
BpipeOutput::init_bpipe_output( std::list<DataFieldPtr>& data_fields ) {

    TraceFct tf( __func__ );

    //
    // delete any node which is not the present in all the input streams.
    //
    auto num = input.get_num();

    data_fields.remove_if( [num](DataFieldPtr& field) { return field->get_count() != num; } );

    //
    //  add the data packet fields to the output bpipe object
    //
    auto died =
        find_if_not( data_fields.begin(),
                     data_fields.end(),
                     [this](DataFieldPtr& field ) {
                         return
                             BPNOERROR
                             == bpipe_dpktf_add( this->bpipe,
                                                 field->get_name().c_str(),
                                                 field->get_type(),
                                                 field->get_matrix()
                                                 );
                     } );
    
    if ( died != data_fields.end() )
        tf.die ( "Unable to add data field" );
  
    bpo = bpipe_output( bpipe, output.c_str( ) );
    if ( NULL == bpo ) 
        tf.die ( "Unable to allocate memory for bpo" );

    // map data field fields
    if ( NULL == ( data = bpipe_map_alloc( bpipe, 1, NULL ) ) && bpipe_errno != BPNOERROR )
        tf.die( std::string("unable to map data fields: ") + bpipe_strerror( bpipe_errno ) );

    /* output header. must do this after calling bpipe_map */
    if ( bpipe_write_hdr( bpipe ) )
        tf.die( std::string("unable to write bpipe header: ") + bpipe_strerror( bpipe_errno ) );
}

void
BpipeOutput::select_data_fields( BPipe* bpipe_input,
                                 const std::vector<std::string>& xfer_fields,
                                 std::list<DataFieldPtr>& data_fields ) {

    TraceFct tf( __func__ );
    void*  last = NULL;     // initialize it for the first time through
    DpktField *dpktf;

    // For each input bpipe, loop thru the 
    while ( (dpktf = bpipe_dpktf_next( bpipe_input, &last )) ) {

        const char* name = bpipe_dpktf_name( dpktf );

        for ( auto& field : xfer_fields ) { 

            if ( field == name )
                continue;

            BPDataType type = bpipe_dpktf_type( dpktf );

            BPMatrix* dpkt_matrix = bpipe_dpktf_matrix( dpktf, BPDSite_CORE, NULL );
            if ( NULL == dpkt_matrix )
                tf.die( std::string("error allocating dpkt matrix: ") + bpipe_strerror( bpipe_errno )  );

            select_data_fields( name, type, dpkt_matrix, data_fields );
        }
    }
}


//
// If a data field is not within the linklist data_fields then
// insert it, else update to take the minimum size of the matrix.
//
void
BpipeOutput::select_data_fields( const char* name, BPDataType t,  BPMatrix* matrix,
                                 std::list<DataFieldPtr>& data_fields ) {

    TraceFct tf( __func__ );

    std::unique_ptr<DataField> datafield( new DataField( name, t, matrix, input.get_num() ) );
    if ( ! datafield )
        TraceFct( __func__ ).die( "error allocating DataField object" );

    auto found = find_if( data_fields.begin(), data_fields.end(),
                          [name]( DataFieldPtr& field ){ return field->get_name() == name; } );

    if ( found == data_fields.end() ) {

        data_fields.emplace_back( std::move(datafield) );

    } else {
        (*found)->update( datafield.get() );
        bpipe_matrix_delete( matrix );
    }
}

void
BpipeOutput::write_dpkts() {

    for ( size_t i = 0; i < input.get_num( ); i++ )
        while ( copy_data( i ) )
            bpipe_write_dpkt( bpipe, data, bpo );

}


void
BpipeOutput::set_output_data_fields( std::list<DataFieldPtr>& data_fields ) {

    using iter_d = std::list<DataFieldPtr>::iterator;

    output_data_fields.clear();
    output_data_fields.insert( output_data_fields.begin(),
                               std::move_iterator<iter_d>( data_fields.begin() ),
                               std::move_iterator<iter_d>( data_fields.end() )
                               );
}
